@@ -1,31 +1,26 @@
 import asyncio
 import base64
 import copy
-from datetime import datetime
-from typing import Dict, List, Union
+from datetime import datetime, timezone
+from typing import Dict, List, Optional, Union, cast
 from urllib.parse import urlparse
 
-import poetry_version
-from playwright.async_api import BrowserContext, CDPSession, Page, Request, Response
+from playwright.async_api import BrowserContext, Page, Request, Response
 
 from . import dataclasses
+from .constants import CREATOR_NAME, CREATOR_VERSION, FALLBACK_HTTP_VERSION, HAR_VERSION
 from .utils import (
+    calculate_request_body_size,
+    calculate_request_headers_size,
+    calculate_response_headers_size,
     cookies_for_har,
     datetime_to_millis,
     dict_to_headers,
     millis_to_roundish_millis,
     post_data_for_har,
     query_to_query_params,
-    set_remote_ip_address_as_comment,
 )
 
-__version__ = poetry_version.extract(source_file=__file__)
-
-
-HAR_VERSION: str = "1.2"
-CREATOR_NAME: str = "playwright-har-tracer"
-CREATOR_VERSION: str = __version__
-
 
 class HarTracer:
     def __init__(
@@ -60,8 +55,6 @@ class HarTracer:
             entries=[],
         )
 
-        self._response_received_events: List[dataclasses.cdp.ResponseReceivedEvent] = []
-
         context.on("page", self.on_page)
 
     def on_request(self, page: Page, request: Request) -> None:
@@ -73,33 +66,33 @@ class HarTracer:
 
         har_entry = dataclasses.har.Entry(
             pageref=page_entry.id,
-            started_date_time=datetime.now(),
+            started_date_time=datetime.now(timezone.utc),
             time=-1,
             request=dataclasses.har.Request(
                 method=request.method,
                 url=request.url,
-                http_version="HTTP/1.1",
+                http_version=FALLBACK_HTTP_VERSION,
                 cookies=[],
                 headers=[],
                 query_string=query_to_query_params(parsed_url.query),
-                post_data=None,
+                post_data=post_data_for_har(request),
                 headers_size=-1,
-                body_size=-1,
+                body_size=calculate_request_body_size(request) or 0,
             ),
             response=dataclasses.har.Response(
                 status=-1,
                 status_text="",
-                http_version="HTTP/1.1",
+                http_version=FALLBACK_HTTP_VERSION,
                 cookies=[],
                 headers=[],
                 content=dataclasses.har.Content(
                     size=-1,
-                    mime_type=request.headers.get("content-type")
-                    or "application/octet-stream",
+                    mime_type=request.headers.get("content-type") or "x-unknown",
                 ),
                 headers_size=-1,
                 body_size=-1,
                 redirect_url="",
+                _transfer_size=-1,
             ),
             cache=dataclasses.har.Cache(before_request=None, after_request=None),
             timings=dataclasses.har.Timings(send=-1, wait=-1, receive=-1),
@@ -132,18 +125,17 @@ class HarTracer:
         har_entry.response = dataclasses.har.Response(
             status=response.status,
             status_text=response.status_text,
-            http_version="HTTP/1.1",
+            http_version=FALLBACK_HTTP_VERSION,
             cookies=cookies_for_har(response.headers.get("set-cookie"), "\n"),
             headers=dict_to_headers(response.headers),
             content=dataclasses.har.Content(
                 size=-1,
-                mime_type=response.headers.get(
-                    "content-type", "application/octet-stream"
-                ),
+                mime_type=response.headers.get("content-type", "x-unknown"),
             ),
             headers_size=-1,
             body_size=-1,
             redirect_url="",
+            _transfer_size=-1,
         )
 
         timing = response.request.timing
@@ -201,6 +193,32 @@ class HarTracer:
         )
         har_entry.time = sum([dns, connect, ssl, wait, receive])
 
+        # set server IP address and port
+        async def set_server_ip_and_port():
+            server = cast(
+                Optional[Dict[str, Union[str, int]]], await response.server_addr()
+            )
+            if server is not None:
+                har_entry.server_ip_address = cast(
+                    Optional[str], server.get("ipAddress")
+                )
+                har_entry._server_port = cast(Optional[int], server.get("port"))
+
+        self._tasks.append(self._loop.create_task(set_server_ip_and_port()))
+
+        # set security details
+        async def set_security_details():
+            security_details = cast(
+                Optional[Dict[str, Union[str, int, float]]],
+                await response.security_details(),
+            )
+            if security_details is not None:
+                har_entry._security_details = dataclasses.har.SecurityDetails.from_dict(
+                    security_details
+                )
+
+        self._tasks.append(self._loop.create_task(set_security_details()))
+
         if self._omit_content is False and response.status == 200:
 
             async def on_response_task():
@@ -210,9 +228,37 @@ class HarTracer:
 
             self._tasks.append(self._loop.create_task(on_response_task()))
 
+    def on_request_finished(self, page: Page, request: Request):
+        har_entry = self._entries.get(request)
+        if har_entry is None:
+            return
+
+        async def handle_finished_request():
+            response = await request.response()
+            if response is None:
+                return
+
+            # TODO
+            http_version = FALLBACK_HTTP_VERSION
+            transfer_size = -1
+            headers_size = calculate_response_headers_size(
+                http_version, response.status, response.status_text, response.headers
+            )
+            body_size = -1
+
+            har_entry.request.http_version = http_version
+            har_entry.response.body_size = body_size
+            har_entry.response.headers_size = headers_size
+            har_entry.response._transfer_size = transfer_size
+            har_entry.request.headers_size = calculate_request_headers_size(
+                request.method, request.url, http_version, request.headers
+            )
+
+        self._tasks.append(self._loop.create_task(handle_finished_request()))
+
     def on_page(self, page: Page) -> None:
         page_entry = dataclasses.har.Page(
-            started_date_time=datetime.now(),
+            started_date_time=datetime.now(timezone.utc),
             id=f"page_{self._last_page}",
             title="",
             page_timings=dataclasses.har.PageTimings(on_content_load=-1, on_load=-1),
@@ -223,6 +269,9 @@ class HarTracer:
         self._log.pages.append(page_entry)
 
         page.on("request", lambda request: self.on_request(page, request))
+        page.on(
+            "requestfinished", lambda request: self.on_request_finished(page, request)
+        )
         page.on("response", lambda response: self.on_response(page, response))
 
         def on_dom_content_loaded(page: Page) -> None:
@@ -270,15 +319,6 @@ class HarTracer:
         page.on("domcontentloaded", lambda: on_dom_content_loaded(page))
         page.on("load", lambda: on_load(page))
 
-    async def enable_response_received_event_tracing(self, client: CDPSession):
-        await client.send("Network.enable")
-        client.on(
-            "Network.responseReceived",
-            lambda data: self._response_received_events.append(
-                dataclasses.cdp.ResponseReceivedEvent.from_dict(data)
-            ),
-        )
-
     async def flush(self) -> dataclasses.har.Har:
         await asyncio.gather(*self._tasks)
 
@@ -301,5 +341,4 @@ class HarTracer:
 
         log = copy.deepcopy(self._log)
         har = dataclasses.har.Har(log=log)
-        har = set_remote_ip_address_as_comment(har, self._response_received_events)
         return har
